#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>
#include <time.h>
#include <pthread.h>
#include <semaphore.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include "fcgi_stdio.h"
#include "libpq-fe.h"
#include "libxml/parser.h"
#include "libxml/tree.h"
#include "ldap.h"
#include "inno-lib/inno-lib.h"
#include "reporting/reporting_config.h"
#include "cdr.h"
#include "pg_report.h"

#define DECORATED_NUMBER_LEN        64

#define COMMA_CHAR                  0x2C
#define FORWARDSLASH_CHAR           0x2F
#define LEFT_CURLY_BRACKET_CHAR     0x7B
#define RIGHT_CURLY_BRACKET_CHAR    0x7D

typedef struct _ldap_conn_ {
    LDAP *ld;
    ldap_t **ldap_server;
} ldap_conn_t;


int rebindproc(LDAP *ld, LDAP_CONST char *url, ber_tag_t request, ber_int_t msgid, void *params)
{
    ldap_t *ldap_server = (ldap_t *)params;
    int ldap_ret;
    char *ldap_error;

    if(url) {
        do_log(LOG_REPORTING, "", "cdr.fcgi: Rebind function called(%d) = %d %s", request, msgid, url);
    }
    else {
        do_log(LOG_REPORTING, "", "cdr.fcgi: Rebind function called(%d) = %d Url empty!", request, msgid);
    }

    if ((ldap_ret = ldap_simple_bind_s(ld, ldap_server->user, ldap_server->password)) != LDAP_SUCCESS) {
        ldap_error = ldap_err2string(ldap_ret);
        do_log(LOG_REPORTING, "", "ldap_rebind error(%d): %s", ldap_ret, ldap_error);
	ldap_unbind_s(ld);

        return LDAP_REFERRAL;
    }
    return LDAP_SUCCESS;
}

static int
lookup_msg(char msg[])
{
    int i, msg_len;
    const int available_msgs = num_msg_lookup_table;
    char *msg_lower;
   
    msg_len = strlen(msg);
    msg_lower = (char *)malloc(msg_len+1);
    for(i=0; i<msg_len; i++) {
        msg_lower[i] = tolower(msg[i]);
    }
    msg_lower[i] = '\0';
    for(i=0; i<available_msgs; i++) {
        if( !strcmp( msg_lower, msg_lookup_table[i]) ) {
            free(msg_lower);
            return i;
        }
    }
    free(msg_lower);
    return UNKNOWN_MSG;
}

char * inno_escape(char * inp)
{
    if(!inp) return NULL;

    int i, ind_out;
    int len = strlen(inp);
    char * out = malloc(len * 2 * sizeof(char) + 1);

    out[0] = 0;

    for(i=0,ind_out=0;i<len;i++) {
        switch((int)inp[i]) {
            case COMMA_CHAR:                /*,*/
                out[ind_out++] = FORWARDSLASH_CHAR;
                out[ind_out++] = COMMA_CHAR;
                break;
            case FORWARDSLASH_CHAR:         /*/*/
                out[ind_out++] = FORWARDSLASH_CHAR;
                out[ind_out++] = FORWARDSLASH_CHAR;
                break;
            case LEFT_CURLY_BRACKET_CHAR:   /*{*/
                out[ind_out++] = FORWARDSLASH_CHAR;
                out[ind_out++] = LEFT_CURLY_BRACKET_CHAR;
                break;
            case RIGHT_CURLY_BRACKET_CHAR:  /*}*/
                out[ind_out++] = FORWARDSLASH_CHAR;
                out[ind_out++] = RIGHT_CURLY_BRACKET_CHAR;
                break;
            default:
                out[ind_out++] = inp[i];
        }
    }
    out[ind_out++] = 0;

    return out;
}

static size_t http_callback(void *ptr, size_t size, size_t nmemb, void *data)
{
    size_t realsize = (size * nmemb);
    http_req *mem = (http_req *)data;
    char *tmp = (char *)ptr;

    if((realsize+1) > HTTP_MAX_SIZE_DATA) return realsize;

    tmp[realsize] = 0;

    memcpy(&(mem->buf[0]), ptr, realsize+1);
    mem->size = realsize+1;

    return realsize;
}

char * alarm_code(int error) 
{
    char code[32];
    int alarm_len = sprintf(code,"0x00180");
    if(error<0x10) sprintf(&code[alarm_len],"00%x",error);
    else if(error<0x100) sprintf(&code[alarm_len],"0%x",error);
    else sprintf(&code[alarm_len],"%x",error);
    return str_malloc(code);
}

char * alarm_file(int error) 
{
    char * code = alarm_code(error);
    char * f = (char *)malloc(strlen(code) + strlen(ALARM_FILE_PREFIX) + 1);
    sprintf(&f[0], "%s%s", ALARM_FILE_PREFIX, code);
    free(code);
    return f;
}

void send_alarm(FCGX_Request *cgi, char * text, char * msg, int error, int clear, int event)
{
    char * f = alarm_file(error);
    FILE * fp;
    if(!clear && !event && (fp = fopen(f, "r")) != NULL) {
        fclose(fp);
        free(f);
        return;
    }
    inno_config_t * cfg = read_inno_config();
    if(cfg->error) {
        do_log(LOG_REPORTING, "", "cdr.fcgi: reading innovaphone config failed: %s", cfg->error);
        cleanup_inno_config(cfg);
        free(f);
        return;
    }
    CURL *curl;
    CURLcode res;
    http_req http_data;
    char post_data[8192];
    time_t datetime = time (NULL);
    char * protocol = "http", * host = "LAP", * pbx = NULL, * https;
    char * code;
    int ssl = 0;
    if (cgi) {
        host = FCGX_GetParam("HTTP_HOST", cgi->envp);
    }
    if(cfg->alarm->ip && *cfg->alarm->ip) {
        pbx = cfg->alarm->ip;
        if(cfg->alarm->https > 0) {
            ssl = 1;
            protocol = "https";
        }
    }
    else if(cgi) {
        pbx = FCGX_GetParam("REMOTE_ADDR", cgi->envp);
        if (!strncmp("::ffff:", pbx, 7)) {  // IPV6 address followed by an IPV4 address
            pbx += 7;
        }
        https = FCGX_GetParam("HTTPS", cgi->envp);
        if(https && (!strcmp(https, "on") || !strcmp(https, "ON"))) {
            protocol = "https";
            ssl = 1;
        }
    }
    if (!pbx) {
        do_log(LOG_REPORTING, "", "couldn't send alarm as no alarm server could be determined!");
        cleanup_inno_config(cfg);
        free(f);
        return;
    }
    char * xml_text = xml_entities(text);
    char * xml_msg = xml_entities(msg);

    code = alarm_code(error);
    char * fault_type = "";
    if (event == 1) {
        fault_type = "error";
    }
    else if (clear == 1) {
        fault_type = "clear";
    }
    else {
        fault_type = "alarm";
    }
    sprintf(post_data, "_fault_xml_post <fault type=\"%s\" severity=\"1\" active=\"false\" code=\"%s\" date=\"%d\" host=\"%s\", xsl=\"fault_reporting.xsl\"><text>%s</text><source>LINUX-Reporting</source><details><cdr>%s</cdr></details></fault>", 
        fault_type,
        code, 
        (int)datetime, 
        host, 
        xml_text, 
        xml_msg);
    free(code);
    free(xml_text);
    free(xml_msg);
    char url[250];
    if(cfg->alarm->port && *cfg->alarm->port) {
        sprintf(url,  "%s://%s:%s/LOG0/FAULT/post-msg", protocol, pbx, cfg->alarm->port);
    }
    else {
        sprintf(url,  "%s://%s/LOG0/FAULT/post-msg", protocol, pbx);
    }
    if(curl_global_init(CURL_GLOBAL_ALL) != 0) {
        do_log(LOG_REPORTING, "", "cdr.fcgi: global curl init for PBX alarm failed");
        cleanup_inno_config(cfg);
        free(f);
        return;
    }

    curl = curl_easy_init();
    if(!curl) {
        do_log(LOG_REPORTING, "", "cdr.fcgi: curl init for PBX alarm failed");
        cleanup_inno_config(cfg);
        free(f);
        return;
    }

    curl_easy_setopt(curl, (CURLoption)CURLOPT_URL, url);
    curl_easy_setopt(curl, (CURLoption)CURLOPT_TIMEOUT, 1);
    curl_easy_setopt(curl, (CURLoption)CURLOPT_WRITEFUNCTION, http_callback);
    curl_easy_setopt(curl, (CURLoption)CURLOPT_WRITEDATA, (void *)&http_data);
    curl_easy_setopt(curl, (CURLoption)CURLOPT_USERAGENT, "libcurl-agent/1.0");
    curl_easy_setopt(curl, (CURLoption)CURLOPT_POST, 1);
    curl_easy_setopt(curl, (CURLoption)CURLOPT_POSTFIELDS, post_data);
    curl_easy_setopt(curl, (CURLoption)CURLOPT_SSL_VERIFYPEER, 0);
    curl_easy_setopt(curl, (CURLoption)CURLOPT_SSL_VERIFYHOST, 0);
    if(cfg->alarm->ip && *cfg->alarm->ip && cfg->alarm->user && *cfg->alarm->user) {
        curl_easy_setopt(curl, (CURLoption)CURLOPT_USERNAME, cfg->alarm->user);
        curl_easy_setopt(curl, (CURLoption)CURLOPT_PASSWORD, cfg->alarm->password);
        curl_easy_setopt(curl, (CURLoption)CURLOPT_HTTPAUTH, CURLAUTH_ANY);
    }
    if((res = curl_easy_perform(curl)) != 0) {
        do_log(LOG_REPORTING, "", "cdr.fcgi: ERROR in curl easy perform: %d, url: %s", res, url);
        curl_easy_cleanup(curl);
        curl_global_cleanup();
        cleanup_inno_config(cfg);
        free(f);
        return;
    }

    curl_easy_cleanup(curl);
    if(http_data.buf[0] == '?') {
        do_log(LOG_REPORTING, "", "cdr.fcgi: alarm set at PBX failed");
    }
    else {
        if(clear == 1) {
            unlink(f);
        }
        else if(event == 0) {   // only write alarm file for alarms, not for events
            FILE * fp;
            if((fp = fopen(f, "w")) != NULL) {
                fclose(fp);
            }
        }
    }
    free(f);
    curl_global_cleanup();
    cleanup_inno_config(cfg);
}

void clear_alarm(FCGX_Request *cgi, int error) 
{
    char * f = alarm_file(error);
    FILE * fp;
    if((fp = fopen(f, "r")) != NULL) {
        fclose(fp);
        send_alarm(cgi, "", "", error, 1, 0);
    }
    free(f);
}

void free_cdr_data(struct _cdr *cdr_data)
{
    int i, j, k;
    struct _cdrs_info_ * cdr;

    if(cdr_data->cdrs) {
        if(trace_level & TRACE_STD) {
            do_log(LOG_REPORTING, "", "free cdr data %d", cdr_data->num_cdrs);
        }
        if(cdr_data->cdrs) {
            for(i=0; i<cdr_data->num_cdrs; i++) {
                if((cdr = cdr_data->cdrs[i]) != NULL) {
                    // events
                    if(cdr->events_values) {
                        for(j=0; j<cdr->num_events; j++) {
                            for(k=0; k<EVENTS_NUM_FIELDS; k++) { 
                                if(cdr->events_values[j*EVENTS_NUM_FIELDS+k]) {
                                    //do_log(LOG_REPORTING, "", "free ev %d %d", j, k);
                                    free(cdr->events_values[j*EVENTS_NUM_FIELDS+k]);
                                }
                            }
                        }
                        free(cdr->events_values);
                    }
                    if(cdr->ev_dn_escaped) {
                        for(j=0; j<cdr->num_events; j++) {
                            if(cdr->ev_dn_escaped[j]) {
                                //do_log(LOG_REPORTING, "", "free dn escaped %d", j);
                                free(cdr->ev_dn_escaped[j]);
                            }
                        }
                        free(cdr->ev_dn_escaped);
                    }
                    if(cdr->ev_h323_escaped) {
                        for(j=0; j<cdr->num_events; j++) {
                            if(cdr->ev_h323_escaped[j]) {
                                //do_log(LOG_REPORTING, "", "free h323 escaped %d", j);
                                free(cdr->ev_h323_escaped[j]);
                            }
                        }
                        free(cdr->ev_h323_escaped);
                    }
                    if(cdr->search_ldap) free(cdr->search_ldap);
                    if(cdr->ev_cause) free(cdr->ev_cause);
                    if(cdr->ev_time) free(cdr->ev_time);
                    if(cdr->events_query) free(cdr->events_query);
                    // groups
                    if(cdr->groups_name_escaped) {
                        for(j=0; j<cdr->num_groups; j++) {
                            if(cdr->groups_name_escaped[j]) {
                                //do_log(LOG_REPORTING, "", "free grp escaped %d", j);
                                free(cdr->groups_name_escaped[j]);
                            }
                        }
                        free(cdr->groups_name_escaped);
                    }
                    if(cdr->groups_values) {
                        for(j=0; j<cdr->num_groups; j++) {
                            for(k=0; k<GROUPS_NUM_FIELDS; k++) { 
                                 if(cdr->groups_values[j*GROUPS_NUM_FIELDS+k]) {
                                     //do_log(LOG_REPORTING, "", "free grp %d %d", j, k);
                                     free(cdr->groups_values[j*GROUPS_NUM_FIELDS+k]);
                                 }
                            }
                        }
                        free(cdr->groups_values);
                    }
                    if(cdr->groups_query) free(cdr->groups_query);
                    // cdr
                    for(j=0; j<CDR_NUM_FIELDS; j++) {
                        if(cdr->cdr_values[j]) {
                            //do_log(LOG_REPORTING, "", "free cdr %d", j);
                            free(cdr->cdr_values[j]);
                        }
                        cdr->cdr_values[j] = NULL;
                    }
                    // cdrp
                    for(j=0; j<CDRP_NUM_FIELDS; j++) {
                        if(cdr->cdrp_values[j]) {
                            //do_log(LOG_REPORTING, "", "free cdrp %d", j);
                            free(cdr->cdrp_values[j]);
                        }
                        cdr->cdrp_values[j] = NULL;
                    }
                    if(cdr->cdrp_flow) free(cdr->cdrp_flow);
                    if(cdr->cdrp_groups) free(cdr->cdrp_groups);

                    free(cdr);
                    cdr = NULL;
                }
            }
            free(cdr_data->cdrs);
        }
        cdr_data->cdrs = NULL;
        cdr_data->num_cdrs = 0;
    }
    if(trace_level & TRACE_STD) {
        do_log(LOG_REPORTING, "", "cdr data freed");
    }
}

char * xml_get_prop(xmlNodePtr node, char * from) {

    char *ret = (char *)xmlGetProp(node, (xmlChar *)from); 

    if(!ret) 
        return str_malloc("");
    return ret;
}

static void defaultNoticeProcessor (void * arg, const char * message) 
{ 
    //syslog(LOG_INFO, "CDR DB ERROR: %s", message);
}

int error(FCGX_Request *cgi, PGconn * conn, char * msg, char * query, char * xml, int error, int rollback, xmlDocPtr doc, int return_error, int alarm, int event) {
    char * pg_msg = (conn ? PQerrorMessage(conn) : "");
    int len = (strlen(pg_msg) + strlen(msg) + strlen(query) + 1) * sizeof(char);
    char * buffer = malloc(len);
    snprintf(buffer, len, "%s%s %s", msg, pg_msg, query);
    do_log(LOG_REPORTING, "", "%s", buffer);
    do_log(LOG_REPORTING, "", "%s", xml);
    if(cgi && alarm) {
        send_alarm(cgi, buffer, xml, error, 0, event);
    }
    if(conn) {
        if(rollback) {
            PQexec(conn, "ROLLBACK");
        }
        PQfinish(conn);
    }
    if(doc)
        xmlFreeDoc(doc);
    free(buffer);
    if(return_error) {
        return error;
    }
    else {
        return 0;
    }
}

static char * ldap_rem_chars(char *in) {
    int i, ind_out = 0;
    char *res_number = (char *)malloc(strlen(in)+1);

    for(i=0;i<strlen(in);i++) {
        switch((int)in[i]) {
            case 45:    /*-*/
            case 32:    /* */
            case 40:    /*(*/
            case 41:    /*)*/
                break;
            default:
                res_number[ind_out++] = in[i];
        }
    }
    res_number[ind_out] = '\0';

    return res_number;
}

static char * ldap_parse_name(char *in) {
    int i = 0, ind_out = 0, len = strlen(in);
    char *res_name = (char *)malloc((strlen(in)<<1)+1);

    while ( ((in[i] == 0x3B) || (in[i] == 0x20)) && (i < len) ) i++;

    for(;i<len;) {
        switch((int)in[i]) {
            case COMMA_CHAR:                /*,*/
                res_name[ind_out++] = FORWARDSLASH_CHAR;
                res_name[ind_out++] = COMMA_CHAR;
                break;
            case FORWARDSLASH_CHAR:         /*/*/
                res_name[ind_out++] = FORWARDSLASH_CHAR;
                res_name[ind_out++] = FORWARDSLASH_CHAR;
                break;
            case LEFT_CURLY_BRACKET_CHAR:   /*{*/
                res_name[ind_out++] = FORWARDSLASH_CHAR;
                res_name[ind_out++] = LEFT_CURLY_BRACKET_CHAR;
                break;
            case RIGHT_CURLY_BRACKET_CHAR:  /*}*/
                res_name[ind_out++] = FORWARDSLASH_CHAR;
                res_name[ind_out++] = RIGHT_CURLY_BRACKET_CHAR;
                break;
            default:
                res_name[ind_out++] = in[i];
        }
        i++;
    }
    res_name[ind_out] = '\0';

    return res_name;
}

int ldap_connect(ldap_conn_t *ldap, ldap_t *ldap_server) 
{
    int ldap_ret;
    char *ldap_error;
    int auth_method = LDAP_AUTH_SIMPLE;
    int desired_version = LDAP_VERSION3;
    int cert_flag = LDAP_OPT_X_TLS_NEVER;
    struct timeval bind_tv;

    if ((ldap_ret = ldap_initialize(&ldap->ld, ldap_server->server)) != LDAP_SUCCESS) {
        ldap_error = ldap_err2string(ldap_ret);
        do_log(LOG_REPORTING, "", "ldap_init failed %s error: %s", ldap_server->server, ldap_error);

        return 0;
    }

    if ((ldap_ret = ldap_set_option(ldap->ld, LDAP_OPT_PROTOCOL_VERSION, &desired_version)) != LDAP_OPT_SUCCESS) {
        ldap_error = ldap_err2string(ldap_ret);
        do_log(LOG_REPORTING, "", "ldap_set_option failed: %s", ldap_error);

        return 0;
    }

    if(strstr(ldap_server->server, "ldaps") != NULL) {
        if ((ldap_ret = ldap_set_option(NULL, LDAP_OPT_X_TLS_REQUIRE_CERT, &cert_flag)) != LDAP_OPT_SUCCESS) {
            ldap_error = ldap_err2string(ldap_ret);
            do_log(LOG_REPORTING, "", "ldap_set_option LDAP_OPT_X_TLS_REQUIRE_CERT failed: %s", ldap_error);

            return 0;
        }
    }

    memset(&bind_tv, 0x0, sizeof(struct timeval));
    bind_tv.tv_sec = 4;
    bind_tv.tv_usec = 0;
    if ((ldap_ret = ldap_set_option(ldap->ld, LDAP_OPT_TIMEOUT, &bind_tv)) != LDAP_OPT_SUCCESS) {
        ldap_error = ldap_err2string(ldap_ret);
        do_log(LOG_REPORTING, "", "ldap_set_option failed: %s", ldap_error);

        return 0;
    }

    ldap_set_rebind_proc(ldap->ld, rebindproc, (void *)ldap_server);

#if 0 // To be tested, ldap_bind_s is deprecated. That is why we get the warning
    struct berval cred;
    cred.bv_val = strdup(ldap_server->password);
    cred.bv_len = strlen(ldap_server->password);
    if ((ldap_ret = ldap_sasl_bind_s(ldap->ld, ldap_server->user, "DIGEST-MD5", &cred, NULL, NULL, NULL)) != LDAP_SUCCESS) {
#else
    if ((ldap_ret = ldap_bind_s(ldap->ld, ldap_server->user, ldap_server->password, auth_method)) != LDAP_SUCCESS) {
#endif
        ldap_error = ldap_err2string(ldap_ret);
        do_log(LOG_REPORTING, "", "ldap_bind error: %s", ldap_error);

        return 0;
    }

    if(trace_level & TRACE_STD) {
        do_log(LOG_REPORTING, "", "ldap connect OK");
    }

    return 1;
}

int ldap_disconnect(ldap_conn_t *ldap) {
    int ldap_ret;
    char *ldap_error;
    
    if((ldap_ret = ldap_unbind_s(ldap->ld)) != 0)
    {
        ldap_error = ldap_err2string(ldap_ret);
        do_log(LOG_REPORTING, "", "ldap_unbind: %s", ldap_error);
        return 0;
    }

    if(trace_level & TRACE_STD) {
        do_log(LOG_REPORTING, "", "ldap disconnect OK");
    }

    return 1;
}

char * ldap_get_name(ldap_conn_t *ldap, char *ev_e164, int internal) 
{
    int i, j, ldap_ret;
    ldap_t **ptr_servers = ldap->ldap_server;
    ldap_t *ldap_server;
    char *server = NULL;
    char *host = NULL;
    char *port = NULL;
    char *user = NULL;
    char *pwd = NULL;
    char *base = NULL;
    char *o_filter = NULL;
    int  o_filter_len;
    char **nameattrib = NULL;
    char **numberattrib = NULL;
    char *cc = NULL;
    char *ac = NULL;
    char *np = NULL;
    char *ip = NULL;
    char *el = NULL;
    char *sn = NULL;
    int ip_len, np_len, cc_len;
    int ac_len, sn_len, el_len;
    int index_dec;
    int start_char;
    int index_check;
    int check_local_len;
    char *ldap_error;
    LDAPMessage *msg = NULL;
    LDAPMessage *entry = NULL;
    BerElement *ber = NULL;
    char *filter = NULL;
    int filter_len;
    char decorated_number[DECORATED_NUMBER_LEN];
    char check_number[64];
    char check_number_local[64];
    char original_number[64];
    int original_set = 0;
    int found_number;
    char *found_name = NULL;
    char *found_names[8];
    int tot_name_length;
    int check_ldap_entries;
    char *phone_number = NULL;
    struct timeval tv;
    int index_server = 0;

    if(trace_level & TRACE_STD) {
        do_log(LOG_REPORTING, "", "ldap get name");
    }

    ldap_server = *ptr_servers;
    while(*ptr_servers != NULL) {
        index_dec = 0;
        start_char = 0;
        index_check = 0;
        check_local_len = 0;
        check_ldap_entries = 0;
        found_number = 0;
        // we check LDAP for internal numbers in some cases.
        if(strncmp(ev_e164, (*ptr_servers)->el, strlen((*ptr_servers)->el)) && (internal != 2)) {
            if(trace_level & TRACE_STD) {
                do_log(LOG_REPORTING, "", "ldap server does not match extern line prefix criteria %s %s [%d]", ev_e164, (*ptr_servers)->el, strlen((*ptr_servers)->el));
            }
            ptr_servers++;
            index_server++;
            continue;
        }
        ldap_server = *ptr_servers++;
        if(trace_level & TRACE_STD) {
            do_log(LOG_REPORTING, "", "Try ldap server[%d] %s", index_server, ldap_server->server);
        }
        index_server++;
        server = ldap_server->server;
        host = ldap_server->host;
        port = ldap_server->port;
        user = ldap_server->user;
        pwd = ldap_server->password;
        base = ldap_server->base;
        o_filter = ldap_server->filter;
        o_filter_len = strlen(o_filter);
        nameattrib = ldap_server->nameattrib;
        numberattrib = ldap_server->numberattrib;
        cc = ldap_server->cc;
        ac = ldap_server->ac;
        np = ldap_server->np;
        ip = ldap_server->ip;
        el = ldap_server->el;
        sn = ldap_server->sn;

        sn_len = strlen(sn);
        el_len = strlen(el);
        ip_len = strlen(ip);
        np_len = strlen(np);
        cc_len = strlen(cc);
        ac_len = strlen(ac);

        if((ev_e164 == NULL) || (strlen(ev_e164) > 32)) {
            do_log(LOG_REPORTING, "", "LDAP: number not allowed for ldap search=%s", ev_e164 ? ev_e164 : "(NULL)");
            
            return NULL;
        }

        if(!el_len || !ip_len || !cc_len) { // Unknown
            if(trace_level & TRACE_STD) {
                do_log(LOG_REPORTING, "", "some of the dialing parameters are not set=%d-%d-%d", el_len, ip_len, cc_len);
            }
            decorated_number[index_dec++] = ev_e164[0];
            check_number[index_check++] = ev_e164[0];
            for(start_char=1; start_char<strlen(ev_e164); start_char++) {
                decorated_number[index_dec++] = 0x2A; // *
                decorated_number[index_dec++] = ev_e164[start_char];
                check_number[index_check++] = ev_e164[start_char];
            }
            check_number[index_check] = '\0';
            decorated_number[index_dec] = '\0';
            if(trace_level & TRACE_STD) {
                do_log(LOG_REPORTING, "", "number=%s-%s-%s", ev_e164, decorated_number, check_number);
            }
        } 
        else if(strncmp(ev_e164, el, el_len) == 0) {  // External Numbers
            decorated_number[index_dec++] = 0x2B;         // +
            check_number[index_check++] = 0x2B;           // +
            if(strncmp(&ev_e164[el_len], ip, ip_len) == 0) {
                if(strlen(ev_e164) == (el_len+ip_len)) {   // wrong dialed, 000 and hang up
                    return NULL;
                } 
                start_char = el_len+ip_len;  // 3
                if((cc_len > 0) && (strncmp(&ev_e164[el_len+ip_len], cc, cc_len) == 0)) {
                    if(strlen(ev_e164) <= (el_len+ip_len+cc_len)) {   // wrong dialed, 00cc and hang up
                        return NULL;
                    } 
                    // National Prefix not present!
                    if((ac_len > 0) && (strncmp(&ev_e164[el_len+ip_len+cc_len], ac, ac_len) == 0)) {
                        if(strlen(ev_e164) <= (el_len+ip_len+cc_len+ac_len)) {   // wrong dialed, 000 and hang up
                            return NULL;
                        } 
                        if(np_len > 0) {
                            memcpy(&check_number_local[0],      np, np_len);
                            memcpy(&check_number_local[np_len], &ev_e164[el_len+ip_len+cc_len+ac_len], (strlen(ev_e164)-(el_len+ip_len+cc_len+ac_len)));
                            check_local_len = np_len+(strlen(ev_e164)-(el_len+ip_len+cc_len+ac_len));
                            check_number_local[check_local_len] = '\0';
                            if(trace_level & TRACE_STD) {
                                do_log(LOG_REPORTING, "", "check local number=%s %s len=%d", ev_e164, check_number_local, check_local_len);
                            }
                        }
                    }
                }
            } else if((np_len > 0) && (strncmp(&ev_e164[el_len], np, np_len) == 0)) { // Maybe there is no national prefix (Norway)
                if(strlen(ev_e164) == (el_len+np_len)) {   // wrong dialed, 00 and hang up
                    return NULL;
                } 
                if((ac_len > 0) && (strncmp(&ev_e164[el_len+np_len], ac, ac_len) == 0)) {
                    if(strlen(ev_e164) <= (el_len+np_len+ac_len)) {   // wrong dialed, 00ac and hang up
                        return NULL;
                    }
                    memcpy(&check_number_local[0],      np, np_len);
                    memcpy(&check_number_local[np_len], &ev_e164[el_len+np_len+ac_len], (strlen(ev_e164)-(el_len+np_len+ac_len)));
                    check_local_len = np_len+(strlen(ev_e164)-(el_len+np_len+ac_len));
                    check_number_local[check_local_len] = '\0';
                    if(trace_level & TRACE_STD) {
                        do_log(LOG_REPORTING, "", "check local number=%s %s len=%d", ev_e164, check_number_local, check_local_len);
                    }
                }
                start_char = el_len+np_len; // 2
                for(i=0;i<cc_len;i++) {
                    decorated_number[index_dec++] = 0x2A; // *
                    decorated_number[index_dec++] = cc[i];
                    check_number[index_check++] = cc[i];
                }
            } else {
                if(strlen(ev_e164) == el_len) {            // wrong dialed, 0 and hang up
                    return NULL;
                } 
                start_char = el_len;  // 1
                for(i=0;i<cc_len;i++) {
                    decorated_number[index_dec++] = 0x2A; // *
                    decorated_number[index_dec++] = cc[i];
                    check_number[index_check++] = cc[i];
                }
                for(i=0;i<ac_len;i++) { // If there is no national prefix, area code should also be empty
                    decorated_number[index_dec++] = 0x2A; // *
                    decorated_number[index_dec++] = ac[i];
                    check_number[index_check++] = ac[i];
                }
            }
            for(;start_char<strlen(ev_e164);start_char++) {
                decorated_number[index_dec++] = 0x2A; // *
                decorated_number[index_dec++] = ev_e164[start_char];
                check_number[index_check++] = ev_e164[start_char];
            }
            for(start_char=0; start_char<strlen(ev_e164); start_char++) {
                original_number[start_char] = ev_e164[start_char];
            }
            check_number[index_check] = '\0';
            decorated_number[index_dec] = '\0';
            original_number[strlen(ev_e164)] = '\0';
            original_set = 1;
            if(trace_level & TRACE_STD) {
                do_log(LOG_REPORTING, "", "external number=%s-%s-%s-%s", ev_e164, decorated_number, check_number, original_number);
            }
        }
        else { // Internal Numbers
            decorated_number[index_dec++] = 0x2B;     // +
            check_number[index_check++] = 0x2B;
            for(i=0;i<cc_len;i++) {
                decorated_number[index_dec++] = 0x2A; // *
                decorated_number[index_dec++] = cc[i];
                check_number[index_check++] = cc[i];
            }
            for(i=0;i<ac_len;i++) {
                decorated_number[index_dec++] = 0x2A; // *
                decorated_number[index_dec++] = ac[i];
                check_number[index_check++] = ac[i];
            }
            for(i=0;i<sn_len;i++) {
                decorated_number[index_dec++] = 0x2A; // *
                decorated_number[index_dec++] = sn[i];
                check_number[index_check++] = sn[i];
            }
            for(i=0;start_char<strlen(ev_e164);start_char++,i++) {
                decorated_number[index_dec++] = 0x2A; // *
                decorated_number[index_dec++] = ev_e164[start_char];
                check_number[index_check++] = ev_e164[start_char];
            }
            check_number[index_check] = '\0';
            decorated_number[index_dec] = '\0';
            if(trace_level & TRACE_STD) {
                do_log(LOG_REPORTING, "", "internal number=%s-%s-%s", ev_e164, decorated_number, check_number);
            }
        }

        filter_len = 8;
        for(i=0;i<MAX_NUM_ATTRIBUTES;i++) {
            if(numberattrib[i]) filter_len += ((strlen(numberattrib[i])+DECORATED_NUMBER_LEN+3)<<1);
        }
        if(o_filter_len) {
            filter_len += o_filter_len+3;  // (&)
        }
        if((filter = (char *)malloc(sizeof(char)*filter_len)) == NULL) {
            return NULL;
        }
        /* build search filter */
        if(o_filter_len) {
            strcpy(filter,"(&");
            strcat(filter, o_filter);
            strcat(filter,"(|");
        }
        else
            strcpy(filter,"(|");
        for(i=0;i<MAX_NUM_ATTRIBUTES;i++) {
            if(numberattrib[i]) {
                strcat(filter,"(");
                strcat(filter,numberattrib[i]);
                strcat(filter,"=");
                strcat(filter, decorated_number);
                strcat(filter,")");
                strcat(filter,"(");
                strcat(filter,numberattrib[i]);
                strcat(filter,"=");
                strcat(filter, check_number);
                strcat(filter,")");
                if(original_set) {
                    strcat(filter,"(");
                    strcat(filter,numberattrib[i]);
                    strcat(filter,"=");
                    strcat(filter, original_number);
                    strcat(filter,")");
                }
            }
        }
        if(o_filter_len)
            strcat(filter, "))");
        else
            strcat(filter, ")");

        if(trace_level & TRACE_STD) {
            do_log(LOG_REPORTING, "", "filter=%s", filter);
            do_log(LOG_REPORTING, "", "conn param=%s-%s-%s-%s-%s", server, user, pwd, base, o_filter);
            do_log(LOG_REPORTING, "", "dialing=%s-%s-%s-%s-%s-%s", cc, ac, np, ip, el, sn);
        }

        if(!ldap_connect(ldap, ldap_server)) {
            free(filter);
            continue;
        }
        
        // double, name+numbers
        char *name_numberattrib[MAX_NUM_ATTRIBUTES<<1];
        for(j=0;j<MAX_NUM_ATTRIBUTES<<1;j++) {
            name_numberattrib[j] = NULL;
        }
        for(j=0,i=0;j<MAX_NUM_ATTRIBUTES;j++) {
            if(nameattrib[j]) {
                name_numberattrib[i++] = nameattrib[j];
            }
        }
        for(j=0;j<MAX_NUM_ATTRIBUTES;j++) {
            if(numberattrib[j]) {
                name_numberattrib[i++] = numberattrib[j];
            }
        }
        memset(&tv, 0x0, sizeof(struct timeval)); 
        tv.tv_sec = 5;
        tv.tv_usec = 0;
        if ((ldap_ret = ldap_search_st(ldap->ld, base, LDAP_SCOPE_SUBTREE, filter, name_numberattrib, 0, &tv, &msg)) != LDAP_SUCCESS) {
            ldap_error = ldap_err2string(ldap_ret);
            do_log(LOG_REPORTING, "", "ldap_search error[%d]: %s", ldap_ret, ldap_error);
            if(ldap_ret == LDAP_OPERATIONS_ERROR) {
                if ((ldap_ret = ldap_search_st(ldap->ld, base, LDAP_SCOPE_SUBTREE, filter, NULL, 0, &tv, &msg)) != LDAP_SUCCESS) {
                    ldap_error = ldap_err2string(ldap_ret);
                    do_log(LOG_REPORTING, "", "ldap_search error[%d]: %s", ldap_ret, ldap_error);
                    free(filter);
                    ldap_disconnect(ldap);
                    continue;
                }
            }
            else {
                free(filter);
                ldap_disconnect(ldap);
                continue;
            }
        }
        free(filter);

        for(j=0;j<MAX_NUM_ATTRIBUTES;j++) {
            found_names[j] = NULL;
        }
        for(entry = ldap_first_entry(ldap->ld, msg); ((entry != NULL) && (check_ldap_entries < 10)); entry = ldap_next_entry(ldap->ld, entry)) {
            char *attr;

            if(trace_level & TRACE_STD) {
                char *dn = NULL;

                if((dn = ldap_get_dn(ldap->ld, entry)) != NULL) {
                    do_log(LOG_REPORTING, "", "Returned dn=%s", dn);
                    ldap_memfree(dn);
                }
            }

            found_number = 0;
            for( attr = ldap_first_attribute(ldap->ld, entry, &ber); attr != NULL; attr = ldap_next_attribute(ldap->ld, entry, ber)) {
                struct berval **vals = NULL;

                if ((vals = ldap_get_values_len(ldap->ld, entry, attr)) != NULL)  {
                    int not_attrib = 1;
                    for(i = 0; vals[i] && vals[i]->bv_len > 0; i++) {
                        for(j=0;(j<MAX_NUM_ATTRIBUTES && not_attrib);j++) {
                            if((nameattrib[j] != NULL) && (!strcasecmp(attr, nameattrib[j]))) {
                                found_names[j] = ldap_parse_name(vals[i]->bv_val);
                                if(trace_level & TRACE_STD) {
                                    do_log(LOG_REPORTING, "", "attrib=%s name=%s", nameattrib[j], found_names[j]);
                                }
                                not_attrib = 0;
                            } else if(!found_number && (numberattrib[j] != NULL) && (!strcasecmp(attr, numberattrib[j]))) {
                                char *ptr_phone;
                                phone_number = ldap_rem_chars(vals[i]->bv_val); 
                                if((ptr_phone = strstr(phone_number, check_number)) != NULL) {
                                    int end_char = (int)(ptr_phone - phone_number) + index_check;

                                    if(trace_level & TRACE_STD) {
                                        do_log(LOG_REPORTING, "", "Dial-Phone %s %d == %d %d %d", ptr_phone, strlen(phone_number), end_char, index_check, (int)(ptr_phone - phone_number));
                                    }

                                    if(strlen(phone_number) == end_char) { 
                                        found_number = 1;
                                    }
                                }
                                if(!found_number && original_set) {
                                    if((ptr_phone = strstr(phone_number, original_number)) != NULL) {
                                        int end_char = (int)(ptr_phone - phone_number) + strlen(ev_e164);

                                        if(trace_level & TRACE_STD) {
                                            do_log(LOG_REPORTING, "", "Original-Phone %s %d == %d %d %d", ptr_phone, strlen(phone_number), end_char, strlen(ev_e164), (int)(ptr_phone - phone_number));
                                        }

                                        if(strlen(phone_number) == end_char) {
                                            found_number = 1;
                                        }
                                    }
                                }
                                if(!found_number && check_local_len) {
                                    if((ptr_phone = strstr(phone_number, check_number_local)) != NULL) {
                                        if(trace_level & TRACE_STD) {
                                            do_log(LOG_REPORTING, "", "Local-Phone %s %d == %d %d", ptr_phone, strlen(phone_number), check_local_len, strlen(ev_e164));
                                        }

                                        if(strlen(phone_number) == check_local_len) {
                                            found_number = 1;
                                        }
                                    }
                                }
                                if(!found_number && !el_len) {
                                    if(trace_level & TRACE_STD) {
                                        do_log(LOG_REPORTING, "", "Dial Direct number(%d,%d,%d) %s(%d) %s(%d)", el_len, ip_len, np_len, phone_number, strlen(phone_number), check_number, strlen(check_number));
                                    }
                                    char dial_direct_number[64];
                                    int offset = 0;
                                    if(ip_len) {
                                        if(!strncmp(check_number, ip, ip_len)) {
                                            offset = ip_len;
                                        }
                                        if(np_len) {
                                            if(!strncmp(&check_number[offset], np, np_len)) {
                                                offset += np_len;
                                            }
                                            else {
                                                offset = 0;   // ip matches but np does not match, cannot be!
                                            }
                                        }
                                        else {
                                            offset = 0;   // np not defined, cannot be!
                                        }
                                    }
#if 0
                                    // ip not defined, cannot be!
                                    else if(np_len && !strncmp(check_number, np, np_len)) {
                                        offset = np_len;
                                    }
#endif
                                    strcpy(dial_direct_number, &check_number[offset]);
                                    if(trace_level & TRACE_STD) {
                                        do_log(LOG_REPORTING, "", "dial_direct_number(%d) %s %s", offset, dial_direct_number, &check_number[offset]);
                                    }
                                    if(offset) {
                                        if((ptr_phone = strstr(phone_number, dial_direct_number)) != NULL) {
                                            found_number = 1;
                                        }
                                    }
                                }
                                if(trace_level & TRACE_STD) {
                                    if(original_set)
                                        do_log(LOG_REPORTING, "", "Phone[%d]: %s-%s-%s", found_number, phone_number, check_number, original_number);
                                    else
                                        do_log(LOG_REPORTING, "", "Phone[%d]: %s-%s", found_number, phone_number, check_number);
                                }
                                free(phone_number);
                                not_attrib = 0;
                            }
                        }
                    }
                    ldap_value_free_len(vals);
                }
                ldap_memfree(attr);
            }
            if(ber != NULL) ber_free(ber,0);
           
            if(found_number) break;

            for(j=0;j<MAX_NUM_ATTRIBUTES;j++) {
                if(found_names[j] != NULL) {
                    free(found_names[j]);
                    found_names[j] = NULL; 
                } 
            }
            check_ldap_entries++;
        }
        if(msg != NULL) ldap_msgfree(msg); 

        ldap_disconnect(ldap);
        if(found_number) {
            tot_name_length = 0;
            for(j=0;j<MAX_NUM_ATTRIBUTES;j++) {
                if(found_names[j] != NULL) {
                    tot_name_length += (strlen(found_names[j])+1);
                } 
            }
            if(tot_name_length) {
                found_name = (char *)malloc((tot_name_length+1)*sizeof(char));
                strcpy(found_name, "");
                for(j=0;j<MAX_NUM_ATTRIBUTES;j++) {
                    if(found_names[j] != NULL) {
                        strcat(found_name, found_names[j]);
                        strcat(found_name, " ");
                        free(found_names[j]);
                        found_names[j] = NULL;
                    } 
                }
                if(trace_level & TRACE_STD) {
                    do_log(LOG_REPORTING, "", "LDAP Name=%s", found_name);
                }
            }
            return found_name;
        }
    }
    if(trace_level & TRACE_STD) {
        do_log(LOG_REPORTING, "", "ldap get name end");
    }
    return NULL;
}

int ldap_init(ldap_conn_t *ldap, char *phys, char *pbx, ldap_servers_t *ldap_cfg)
{
    int i;
    ldap_t **servers;
    ldap_t **ptr_servers;

    if(trace_level & TRACE_STD) {
        do_log(LOG_REPORTING, "", "ldap init servers=%d cfg=%d", ldap_cfg->num_ldap_servers, (ldap_cfg == NULL));
    }

    if(ldap_cfg == NULL) {
        return 2;   // error
    }

    if(!ldap_cfg->num_ldap_servers) {
        return 0;   // no servers
    }

    ldap->ldap_server = (ldap_t **)malloc((ldap_cfg->num_ldap_servers+1)*sizeof(ldap_t *));
    ptr_servers = ldap->ldap_server;
    for(i=0;i<=ldap_cfg->num_ldap_servers;i++) {
        *ptr_servers++ = NULL;
    }
    ptr_servers = ldap->ldap_server;

    servers = ldap_cfg->ldap;
    while(*servers != NULL) {
        ldap_t *tmp = *servers;
        if(*phys) {
            if(trace_level & TRACE_STD) {
                do_log(LOG_REPORTING, "", "check phys[%s] %s %s", tmp->active, tmp->pbx, phys);
            }
            if(!strcasecmp(tmp->pbx, phys) || !strcmp(tmp->pbx, "%")) {
                if(tmp->server == NULL || tmp->user == NULL || tmp->password == NULL || tmp->cc == NULL ||\
                   tmp->ac == NULL || tmp->np == NULL || tmp->ip == NULL || tmp->el == NULL || tmp->sn == NULL ||\
                   tmp->nameattrib[0] == NULL || tmp->numberattrib[0] == NULL) {
                    do_log(LOG_REPORTING, "", "LDAP: Parameters not parsed %x-%x-%x-%x-%x-%x-%x-%x-%x-%x-%x", 
                                              tmp->server, tmp->user, tmp->password, tmp->cc, tmp->ac, tmp->np, tmp->ip, 
                                              tmp->el, tmp->sn, tmp->nameattrib[0], tmp->numberattrib[0]);
                } else {
                    *ptr_servers++ = tmp;
                    if(trace_level & TRACE_STD) {
                        do_log(LOG_REPORTING, "", "conn param=%s-%s-%s", tmp->server, tmp->user, tmp->base);
                        do_log(LOG_REPORTING, "", "dialing=%s-%s-%s-%s-%s-%s", tmp->cc, tmp->ac, tmp->np, tmp->ip, tmp->el, tmp->sn);
                        do_log(LOG_REPORTING, "", "Name attributes=");
                        for(i=0;i<MAX_NUM_ATTRIBUTES;i++) {
                            if(tmp->nameattrib[i]) {
                                do_log(LOG_REPORTING, "", "%s", tmp->nameattrib[i]);
                            }
                        }
                        do_log(LOG_REPORTING, "", "Number attributes=");
                        for(i=0;i<MAX_NUM_ATTRIBUTES;i++) {
                            if(tmp->numberattrib[i]) {
                                do_log(LOG_REPORTING, "", "%s", tmp->numberattrib[i]);
                            }
                        }
                    }
                }
            }
        } else {
            if(trace_level & TRACE_STD) {
                do_log(LOG_REPORTING, "", "check pbx[%s] %s %s", tmp->active, tmp->pbx, pbx);
            }
            if(!strcasecmp(tmp->pbx, pbx) || !strcmp(tmp->pbx, "%")) {
                if(tmp->server == NULL || tmp->user == NULL || tmp->password == NULL || tmp->cc == NULL ||\
                   tmp->ac == NULL || tmp->np == NULL || tmp->ip == NULL || tmp->el == NULL || tmp->sn == NULL ||\
                   tmp->nameattrib[0] == NULL || tmp->numberattrib[0] == NULL) {
                    do_log(LOG_REPORTING, "", "LDAP: Parameters not parsed %x-%x-%x-%x-%x-%x-%x-%x-%x-%x-%x", 
                                              tmp->server, tmp->user, tmp->password, tmp->cc, tmp->ac, tmp->np, tmp->ip, 
                                              tmp->el, tmp->sn, tmp->nameattrib[0], tmp->numberattrib[0]);
                } else {
                    *ptr_servers++ = tmp;
                    if(trace_level & TRACE_STD) {
                        do_log(LOG_REPORTING, "", "conn param=%s-%s-%s", tmp->server, tmp->user, tmp->base);
                        do_log(LOG_REPORTING, "", "dialing=%s-%s-%s-%s-%s-%s", tmp->cc, tmp->ac, tmp->np, tmp->ip, tmp->el, tmp->sn);
                        do_log(LOG_REPORTING, "", "Name attributes=");
                        for(i=0;i<MAX_NUM_ATTRIBUTES;i++) {
                            if(tmp->nameattrib[i]) {
                                do_log(LOG_REPORTING, "", "%s", tmp->nameattrib[i]);
                            }
                        }
                        do_log(LOG_REPORTING, "", "Number attributes=");
                        for(i=0;i<MAX_NUM_ATTRIBUTES;i++) {
                            if(tmp->numberattrib[i]) {
                                do_log(LOG_REPORTING, "", "%s", tmp->numberattrib[i]);
                            }
                        }
                    }
                }
            }
        }
        servers++;
    }
    ptr_servers = ldap->ldap_server;

    if(*ptr_servers == NULL) {
        if(trace_level & TRACE_STD) {
            do_log(LOG_REPORTING, "", "No Ldap server found or servers not active");
        }
        free(ldap->ldap_server);
        ldap->ldap_server = NULL;
        return 0;   // no server match
    }

    if(trace_level & TRACE_STD) {
        do_log(LOG_REPORTING, "", "ldap init OK");
    }

    return 1;   // all ok and server match
}

int ldap_close(ldap_conn_t *ldap) 
{
    if(ldap->ldap_server) free(ldap->ldap_server);

    if(trace_level & TRACE_STD) {
        do_log(LOG_REPORTING, "", "ldap close OK");
    }

    return 1;
}

int send_cdr_to_db(struct _cdrs_info_ * cdr_data, FCGX_Request *cgi, PGconn * conn)
{
    // calculate call flow and groups size
    int i = 0, cdr_id = 0;
    PGresult * res = NULL;
    struct timeval tvBegin, tvEnd, tvDiff;

    if(trace_level & TRACE_TIME) {
        gettimeofday(&tvBegin, NULL);
    }

    // start transaction
    res = PQexec(conn, "BEGIN");
    if (PQresultStatus(res) != PGRES_COMMAND_OK)
    {
        char* sqlstate = PQresultErrorField(res, PG_DIAG_SQLSTATE);

        if(trace_level & TRACE_STD) {
            char* severity = PQresultErrorField(res, PG_DIAG_SEVERITY);
            char* hint = PQresultErrorField(res, PG_DIAG_MESSAGE_HINT);

            do_log(LOG_REPORTING, "", "sqlstate (%s) severity (%s) hint (%s)", sqlstate, severity, hint);
        }

        PQclear(res);
        // http://www.network-theory.co.uk/docs/postgresql/vol1/PostgreSQLErrorCodes.html#SEC377
        if(sqlstate && (!strcmp(sqlstate, "57P01") || !strcmp(sqlstate, "57P02")) ) { // shutdown
            return error(cgi, conn, "DB BEGIN failed: ", sqlstate, NULL, ERROR_DB_HANDLER_NOT_VALID, 0, NULL, ERROR_EXECUTING_BEGIN_CMD, 1, 1);
        }
        return error(cgi, conn, "DB BEGIN failed: ", "BEGIN transaction", NULL, ERROR_INSERTING_CDR, 0, NULL, ERROR_EXECUTING_BEGIN_CMD, 1, 1);
    } 
    PQclear(res);

    // lock table cdr
    res = PQexecPrepared(conn, PG_CDR_LOCK_PREP_TRANS, 0, NULL, NULL, NULL, 0);
    if (PQresultStatus(res) != PGRES_COMMAND_OK)
    {
        PQclear(res);
        return error(cgi, conn, "DB LOCK failed: ", "LOCK query", NULL, ERROR_INSERTING_CDR, 0, NULL, ERROR_EXECUTING_LOCK_CDR_CMD, 1, 1);
    } 
    PQclear(res);

    res = PQexecPrepared(conn, PG_CDR_GUID_PREP_TRANS, 1, (const char * const*)&cdr_data->cdr_values[CDR_CDR_GUID_INDEX], NULL, NULL, 0);
    if (PQresultStatus(res) != PGRES_TUPLES_OK)
    {
        PQclear(res);
        return error(cgi, conn, "DB SELECT cdr_guid failed:", "CDR GUID query", NULL, ERROR_INSERTING_CDR, 0, NULL, ERROR_EXECUTING_QUERY, 1, 1);
    }

    if(PQntuples(res)) {
        if(trace_level & TRACE_DB) {
            do_log(LOG_REPORTING, "", "CDR already in DB = %s", cdr_data->cdr_values[CDR_CDR_GUID_INDEX]);
        }
        PQclear(res);
        return 0;
    }
    PQclear(res);

    // insert cdr
    res = PQexecPrepared(conn, PG_CDR_PREP_TRANS, CDR_NUM_FIELDS, (const char * const*)cdr_data->cdr_values, NULL, NULL, 0);
    if(PQresultStatus(res) != PGRES_TUPLES_OK) {
        int ret_error;
        int alarm = 1;
        char * pq_error = PQresultErrorField(res, PG_DIAG_SQLSTATE);
        PQclear(res);
        if(pq_error && !strcmp(pq_error, "40P01")) {
            alarm = 0;
        }
        ret_error = error(cgi, conn, "DB INSERT cdr failed: ", PG_CDR_QUERY, NULL, ERROR_INSERTING_CDR, 1, NULL, ERROR_EXECUTING_INSERT_CDR_QUERY, alarm, 1);
        return ret_error;
    }
    cdr_id = atoi(PQgetvalue(res, 0, 0));
    PQclear(res);
    if(trace_level & TRACE_DB) {
        do_log(LOG_REPORTING, "", "cdr_id=%d", cdr_id);
    }
    if(cdr_data->num_events) {
        for(i = 0; i < cdr_data->num_events; i++) {
            sprintf(cdr_data->events_values[i*EVENTS_NUM_FIELDS+EVENT_CDR_ID_INDEX], "%d", cdr_id);
        }
        if(trace_level & TRACE_DB) {
            for(i=0; i<cdr_data->num_events; i++) {
                do_log(LOG_REPORTING, "", "event[%d]=%s %s %s %s %s %s %s %s %s %s", i, 
                                      cdr_data->events_values[i*EVENTS_NUM_FIELDS+0], cdr_data->events_values[i*EVENTS_NUM_FIELDS+1],
                                      cdr_data->events_values[i*EVENTS_NUM_FIELDS+2], cdr_data->events_values[i*EVENTS_NUM_FIELDS+3],
                                      cdr_data->events_values[i*EVENTS_NUM_FIELDS+4], cdr_data->events_values[i*EVENTS_NUM_FIELDS+5],
                                      cdr_data->events_values[i*EVENTS_NUM_FIELDS+6], cdr_data->events_values[i*EVENTS_NUM_FIELDS+7],
                                      cdr_data->events_values[i*EVENTS_NUM_FIELDS+8], cdr_data->events_values[i*EVENTS_NUM_FIELDS+9]);
            }
        }
        if(cdr_data->num_events > 5)
            res = PQexecParams(conn, cdr_data->events_query, cdr_data->num_events*EVENTS_NUM_FIELDS, NULL, (const char * const*)cdr_data->events_values, NULL, NULL, 0);
        else
            res = PQexecPrepared(conn, cdr_data->events_query, cdr_data->num_events*EVENTS_NUM_FIELDS, (const char * const*)cdr_data->events_values, NULL, NULL, 0);
        if(PQresultStatus(res) != PGRES_COMMAND_OK) {
            int ret_error;

            PQclear(res);
            ret_error = error(cgi, conn, "DB INSERT cdr event failed: ", cdr_data->events_query, NULL, ERROR_INSERTING_CDR, 1, NULL, ERROR_EXECUTING_INSERT_EV_QUERY, 1, 1);
            return ret_error;
        }
        PQclear(res);
    }
    if(cdr_data->num_groups) {
        for(i=0; i<cdr_data->num_groups; i++) {
            sprintf(cdr_data->groups_values[i*GROUPS_NUM_FIELDS+GROUP_CDR_ID_INDEX],"%d",cdr_id);
        }
        if(trace_level & TRACE_DB) {
            for(i=0; i<cdr_data->num_groups; i++) {
                do_log(LOG_REPORTING, "", "group[%d]=%s %s %s %s", i,
                                          cdr_data->groups_values[i*GROUPS_NUM_FIELDS+0], cdr_data->groups_values[i*GROUPS_NUM_FIELDS+1],
                                          cdr_data->groups_values[i*GROUPS_NUM_FIELDS+2], cdr_data->groups_values[i*GROUPS_NUM_FIELDS+3]);
            }
        }
        if(cdr_data->num_groups > 5)
            res = PQexecParams(conn, cdr_data->groups_query, cdr_data->num_groups*GROUPS_NUM_FIELDS, NULL, (const char * const*)cdr_data->groups_values, NULL, NULL, 0);
        else
            res = PQexecPrepared(conn, cdr_data->groups_query, cdr_data->num_groups*GROUPS_NUM_FIELDS, (const char * const*)cdr_data->groups_values, NULL, NULL, 0);
        if(PQresultStatus(res) != PGRES_COMMAND_OK) {
            int ret_error;

            PQclear(res);
            ret_error = error(cgi, conn, "DB INSERT cdr groups failed: ", cdr_data->groups_query, NULL, ERROR_INSERTING_CDR, 1, NULL, ERROR_EXECUTING_INSERT_GRP_QUERY, 1, 1);
            return ret_error;
        }
        PQclear(res);
    }

    sprintf(cdr_data->cdrp_values[CDRP_CDR_ID_INDEX],"%d",cdr_id);
    res = PQexecPrepared(conn, PG_CDRP_PREP_TRANS, CDRP_NUM_FIELDS, (const char * const*)cdr_data->cdrp_values, NULL, NULL, 0);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        int ret_error;

        PQclear(res);
        ret_error = error(cgi, conn, "DB INSERT cdr properties failed: ", PG_CDRP_QUERY, NULL, ERROR_INSERTING_CDR, 1, NULL, ERROR_EXECUTING_INSERT_CDRP_QUERY, 1, 1);
        return ret_error;
    }
    PQclear(res);

    // finally commit the changes
    res = PQexec(conn, "COMMIT");
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        PQclear(res);
        return error(cgi, conn, "DB COMMIT failed: ", "COMMIT transaction", NULL, ERROR_INSERTING_CDR, 1, NULL, ERROR_EXECUTING_COMMIT_CMD, 1, 1);
    }
    PQclear(res);

    if(trace_level & TRACE_TIME) {
        gettimeofday(&tvEnd, NULL);
        timeval_subtract(&tvDiff, &tvEnd, &tvBegin);
        do_log(LOG_REPORTING, "", "cdr insert: %ld.%06ld", tvDiff.tv_sec, tvDiff.tv_usec);
    }

    return 0;
}

int init_db_conn(PGconn **conn, FCGX_Request *cgi)
{
    *conn = PQconnectdb(PG_CONN_STR);
    PQsetNoticeProcessor(*conn, defaultNoticeProcessor, NULL);
    PGresult *res;

    if (PQstatus(*conn) != CONNECTION_OK) {
        do_log(LOG_REPORTING, "", "Connection to database failed: %s", PG_CONN_STR);
        PQfinish(*conn);
        *conn = NULL;
        send_alarm(cgi, "Connection to database failed", "", ERROR_CONNECTING_TO_DB, 0, 0);
        return ERROR_CONNECTING_TO_DB;
    }
    clear_alarm(cgi, ERROR_CONNECTING_TO_DB);

    if (PQsetClientEncoding(*conn, "UTF-8") != 0) {
        do_log(LOG_REPORTING, "", "Setting client encoding failed: %s", "UTF-8");
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_SETTING_ENCODING;
    }

    res = PQprepare(*conn, PG_CDR_PREP_TRANS, PG_CDR_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for cdr %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_CDR_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_CDRP_PREP_TRANS, PG_CDRP_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for cdrp %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_CDRP_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_5EVENTS_PREP_TRANS, PG_5EVENTS_PREP_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for events 5 %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_EVENT_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_4EVENTS_PREP_TRANS, PG_4EVENTS_PREP_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for events 4 %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_EVENT_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_3EVENTS_PREP_TRANS, PG_3EVENTS_PREP_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for events 3 %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_EVENT_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_2EVENTS_PREP_TRANS, PG_2EVENTS_PREP_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for events 2 %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_EVENT_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_1EVENTS_PREP_TRANS, PG_1EVENTS_PREP_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for events 1 %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_EVENT_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_5GROUPS_PREP_TRANS, PG_5GROUPS_PREP_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for groups %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_GROUP_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_4GROUPS_PREP_TRANS, PG_4GROUPS_PREP_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for groups %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_GROUP_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_3GROUPS_PREP_TRANS, PG_3GROUPS_PREP_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for groups 3 %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_GROUP_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_2GROUPS_PREP_TRANS, PG_2GROUPS_PREP_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for groups 2 %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_GROUP_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_1GROUPS_PREP_TRANS, PG_1GROUPS_PREP_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for groups 1 %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_GROUP_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_LDAP_ALL_PREP_TRANS, PG_LDAP_ALL_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for ldap all %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_LDAP_ALL_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_LDAP_ACT_PREP_TRANS, PG_LDAP_ACT_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for ldap active %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_LDAP_ACT_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_CDR_GUID_PREP_TRANS, PG_CDR_GUID_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for cdr guid %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_CDR_GUID_TRANSACTION;
    }
    PQclear(res);

    res = PQprepare(*conn, PG_CDR_LOCK_PREP_TRANS, PG_CDR_LOCK_QUERY, 0, NULL);
    if(PQresultStatus(res) != PGRES_COMMAND_OK) {
        do_log(LOG_REPORTING, "", "Error preparing transaction for cdr lock %s", PQerrorMessage(*conn));
        PQclear(res);
        PQfinish(*conn);
        *conn = NULL;
        return ERROR_PREPARING_CDR_LOCK_TRANSACTION;
    }
    PQclear(res);

    return 0;
}

void close_db_conn(PGconn **conn)
{
    PQfinish(*conn); 
}

int parse_cdr(struct _cdr* cdr_data, char *query_string, FCGX_Request *cgi)
{
    int i, j, send_err = 0, event_msg;
    int cdr_index = 0;
    struct timeval tvBegin, tvBegin1, tvEnd, tvDiff;
    int conn_duration = 0;
    int alert_duration = 0;
    int call_duration = 0;
    int billing_duration = 0;
    int cause = 0;
    int cdrp_groups_seperators = 3;
    char **events_values = NULL;
    char **groups_values = NULL;
    int do_ldap = 0;
    ldap_conn_t ldap_conn;
    int cdrp_flow_seperators = 6;   // seperator chars between/around values
    int cdrp_groups_size, cdrp_flow_size;
    char *groups_query = NULL;
    char *events_query = NULL;
    int index_arg;

    memset(&ldap_conn, 0x0, sizeof(ldap_conn_t));
    
    if(trace_level & TRACE_STD) {
        //system("rm -f /tmp/mem.log");
        system("echo 'Test-CDR:' >> /tmp/mem.log");
        system("ps -eo pmem,pcpu,rss,vsize,args | sort -k 1 | grep 'process.fcgi' | grep -v grep >> /tmp/mem.log");
    }

#if 0
    // only root...
    if(!setpriority(PRIO_PROCESS, getpid(),-10))
        syslog(LOG_INFO,"p=%d",getpriority(PRIO_PROCESS, getpid()));
    else
        syslog(LOG_INFO,"cannot be done...");
#endif

    if(trace_level & TRACE_TIME) {
        gettimeofday(&tvBegin, NULL);
    }

    xmlDocPtr doc = xmlReadMemory(query_string, strlen(query_string), "noname.xml", NULL, 0);
    if(doc == NULL) {
        return error(cgi, NULL, "Failed to parse XML", "", query_string, ERROR_CDR_PARSING_CDR_XML, 0, NULL, ERROR_CDR_PARSING_CDR_XML, 1, 1);
    }
    
    xmlNodePtr node_cdr;
    cdr_data->num_cdrs = 0;
    for(node_cdr = doc->children; node_cdr != NULL; node_cdr = node_cdr->next)
    {
        if(xmlStrcasecmp(node_cdr->name, (xmlChar *)"cdr")) continue; // ignore none cdr nodes

        cdr_data->num_cdrs++;
    }

    cdr_data->cdrs = (struct _cdrs_info_ **)malloc(sizeof(struct _cdrs_info_ *)*cdr_data->num_cdrs);
    for(i=0; i<cdr_data->num_cdrs; i++) {
        cdr_data->cdrs[i] = malloc(sizeof(struct _cdrs_info_));
        memset(cdr_data->cdrs[i], 0, sizeof(struct _cdrs_info_));
    }

    for (node_cdr = doc->children; node_cdr != NULL; node_cdr = node_cdr->next)
    {
        struct _cdrs_info_ *cdr = cdr_data->cdrs[cdr_index];

        if (trace_level & TRACE_STD) {
            do_log(LOG_REPORTING, "", "enter proc cdr: %s %s", doc->children->name, node_cdr->name);
        }

        if (xmlStrcasecmp(node_cdr->name, (xmlChar *)"cdr")) continue; // ignore none cdr nodes

        // check reporting flag
        char * licensed = xml_get_prop(node_cdr, "reporting");
        int license_version = -1;
        if (licensed && *licensed) {
            license_version = atoi((char *)licensed);
        }
        xmlFree(licensed);
        if (license_version != 0 && license_version < LICENSE_VERSION) {
            do_log(LOG_REPORTING, "", "unlicensed CDR (%d): %s", license_version, query_string);
            xmlFreeDoc(doc);
            free_cdr_data(cdr_data);
            return ERROR_UNLICENSED_CDR;
        }
        else {
            if (trace_level & TRACE_TIME) {
                gettimeofday(&tvBegin1, NULL);
            }

            if (trace_level & TRACE_STD) {
                do_log(LOG_REPORTING, "", "search for special process");
            }

            if (check_db_restore()) {
                free_cdr_data(cdr_data);
                xmlFreeDoc(doc);
                return error(cgi, NULL, "DB restore running", "", query_string, ERROR_DB_BEING_RESTORED, 0, NULL, ERROR_DB_BEING_RESTORED, 1, 1);
            }

            if (trace_level & TRACE_TIME) {
                gettimeofday(&tvEnd, NULL);
                timeval_subtract(&tvDiff, &tvEnd, &tvBegin1);
                do_log(LOG_REPORTING, "", "procs: %ld.%06ld", tvDiff.tv_sec, tvDiff.tv_usec);
            }
        }

        // calculate event/group count
        xmlNodePtr child;
        xmlNodePtr child_next;
        int num_events = 0;
        int num_groups = 0;
        int cur_event = 0;
        int cur_group = 0;

        for (child = node_cdr->children; child != NULL; child = child->next) {
            if (!xmlStrcasecmp(child->name, (xmlChar *)"event")) {
                char * event_msg = xml_get_prop(child, "msg");
                // ignore info-to/info-from
                if(!strcasecmp(event_msg, "info-to") || !strcasecmp(event_msg, "info-from")) {
                    free(event_msg);
                    continue;
                }
                free(event_msg);
                num_events++;
                xmlNodePtr sub_event;
                for (sub_event = child->children; sub_event != NULL; sub_event = sub_event->next) {
                    if (sub_event->properties && (!xmlStrcasecmp(sub_event->name, (xmlChar *)"ct") || !xmlStrcasecmp(sub_event->name, (xmlChar *)"cf"))) {
                        num_events++;
                    }
                }
            }
            else if (!xmlStrcasecmp(child->name, (xmlChar *)"user")) {
                xmlNodePtr grp_node;
                for (grp_node = child->children; grp_node != NULL; grp_node = grp_node->next) {
                    if (!xmlStrcasecmp(grp_node->name, (xmlChar *)"grp")) {
                        num_groups++;
                    }
                }
            }
        }
        //syslog(LOG_INFO, "groups=%d, events=%d", num_groups, num_events);
        if (license_version == 0) {  // only for myPBX call lists
            cdr->cdr_values[CDR_LICENSED_INDEX] = str_malloc("false");
        }
        else {
            cdr->cdr_values[CDR_LICENSED_INDEX] = str_malloc("true");
        }
        cdr->cdr_values[CDR_CDR_GUID_INDEX] = xml_get_prop(node_cdr, "id");
        cdr->cdr_values[CDR_GUID_INDEX] = xml_get_prop(node_cdr, "guid");
        cdr->cdr_values[CDR_UTC_INDEX] = xml_get_prop(node_cdr, "utc");
        cdr->cdr_values[CDR_LOCAL_INDEX] = xml_get_prop(node_cdr, "local");
        cdr->cdr_values[CDR_EMAIL_INDEX] = xml_get_prop(node_cdr, "email");
        char * not_exist = NULL;
        if (!*(cdr->cdr_values[CDR_CDR_GUID_INDEX])) {
            not_exist = "id";
        }
        else if (!*(cdr->cdr_values[CDR_GUID_INDEX])) {
            not_exist = "guid";
        }
        else if (!*(cdr->cdr_values[CDR_UTC_INDEX])) {
            not_exist = "utc";
        }
        else if (!*(cdr->cdr_values[CDR_LOCAL_INDEX])) {
            not_exist = "local";
        }
        if (not_exist) {
            char buffer[128];
            sprintf(buffer, "Attribute %s not present in CDR", not_exist);
            free_cdr_data(cdr_data);
            return error(cgi, NULL, buffer, "", query_string, ERROR_CDR_ATTRIBUTE_NOT_PRESENT, 0, doc, ERROR_CDR_ATTRIBUTE_NOT_PRESENT, 1, 1);
        }
        
        int conn_time = -1, conn_end_time = -1, alert_time = -1, disc_time = -1, time = 0, setup_event = 0,
            call_time = -1, call_end_time = -1, billing_time = -1, end_billing_time = -1;
        int billing_call = 0, set_conn_time = 1;
        cause = -1;

        // assign further CDR values
        cdr->cdr_values[CDR_SYS_INDEX] = xml_get_prop(node_cdr, "sys");
        cdr->cdr_values[CDR_PBX_INDEX] = xml_get_prop(node_cdr, "pbx");
        cdr->cdr_values[CDR_NODE_INDEX] = xml_get_prop(node_cdr, "node");
        cdr->cdr_values[CDR_DEVICE_INDEX] = xml_get_prop(node_cdr, "device");
        cdr->cdr_values[CDR_CN_INDEX] = xml_get_prop(node_cdr, "cn");
        cdr->cdr_values[CDR_E164_INDEX] = xml_get_prop(node_cdr, "e164");
        cdr->cdr_values[CDR_H323_INDEX] = xml_get_prop(node_cdr, "h323");
        cdr->cdr_values[CDR_DIR_INDEX] = xml_get_prop(node_cdr, "dir");
        cdr->cdr_values[CDR_PHYS_INDEX] = xml_get_prop(node_cdr, "phys");
        cdr->cdr_values[CDR_DN_ESCAPED_INDEX] = NULL;
        cdr->cdr_values[CDR_TRANSFERRED_INDEX] = str_malloc("0");
        cdr->cdr_values[CDR_FORWARDED_INDEX] = str_malloc("0");
        // Check for billing cases, right now, everything but to, from, ct and cf
        if (strcmp(cdr->cdr_values[CDR_DIR_INDEX], "to")) {
            billing_call = 1;
        }

        // events
        events_values = (char **)malloc(num_events*EVENTS_NUM_FIELDS*sizeof(char *));
        memset(events_values, 0x0, (num_events*EVENTS_NUM_FIELDS*sizeof(char *)));
        cdr->ev_dn_escaped = (char **)malloc(num_events*sizeof(char *));
        cdr->ev_h323_escaped = (char **)malloc(num_events*sizeof(char *));
        cdr->search_ldap = (int *)malloc(num_events*sizeof(int));
        cdr->ev_cause = (int *)malloc(num_events*sizeof(int));
        cdr->ev_time = (int *)malloc(num_events*sizeof(int));
        memset(cdr->ev_dn_escaped, 0x0, num_events*sizeof(char *));
        memset(cdr->ev_h323_escaped, 0x0, num_events*sizeof(char *));
        memset(cdr->search_ldap, 0x0, num_events*sizeof(int));
        memset(cdr->ev_cause, 0x0, num_events*sizeof(int));
        // groups
        groups_values = (char **)malloc(num_groups*GROUPS_NUM_FIELDS*sizeof(char *));
        memset(groups_values, 0x0, (num_groups*GROUPS_NUM_FIELDS*sizeof(char *)));
        cdr->groups_name_escaped = (char **)malloc(num_groups*sizeof(char *));
        memset(cdr->groups_name_escaped, 0x0, num_groups*sizeof(char *));

        cdr->cdrp_values[CDRP_TYPE_INDEX] = str_malloc("int");
        cdr->cdrp_values[CDRP_MORE_CALLS_INDEX] = NULL;
        cdr->cdrp_values[CDRP_REJECTED_INDEX] = NULL;
        for(i=0; i<num_events; i++) {
            cdr->ev_time[i] = -1;
        }
        cdrp_groups_size = 3 + cdrp_groups_seperators;

        cdr->events_values = events_values;
        cdr->num_events = num_events;
        cdr->groups_values = groups_values;
        cdr->num_groups = num_groups;

        // process groups and events
        for(child = node_cdr->children; child != NULL; child = child->next) {
            if(!xmlStrcasecmp(child->name, (xmlChar *)"user")) {
                xmlNodePtr grp_node;
                for(grp_node = child->children; grp_node != NULL; grp_node = grp_node->next) {
                    if(!xmlStrcasecmp(grp_node->name, (xmlChar *)"grp")) {
                        groups_values[cur_group*GROUPS_NUM_FIELDS+GROUP_NAME_INDEX] = xml_get_prop(grp_node, "name");

                        char *grp_name_tmp = inno_escape((char *)groups_values[cur_group*GROUPS_NUM_FIELDS+GROUP_NAME_INDEX]);
                        cdr->groups_name_escaped[cur_group] = xml_entities(grp_name_tmp);
                        free(grp_name_tmp);
                        cdrp_groups_size += cdrp_groups_seperators + strlen(cdr->groups_name_escaped[cur_group]);

                        groups_values[cur_group*GROUPS_NUM_FIELDS+GROUP_TYPE_INDEX] = xml_get_prop(grp_node, "dyn");
                        char * mode = xml_get_prop(grp_node, "mode");
                        if(!strcasecmp(mode, "active")) {
                            groups_values[cur_group*GROUPS_NUM_FIELDS+GROUP_ACT_INDEX] = str_malloc("1");
                        }
                        else {
                            groups_values[cur_group*GROUPS_NUM_FIELDS+GROUP_ACT_INDEX] = str_malloc("0");
                        }
                        groups_values[cur_group*GROUPS_NUM_FIELDS+GROUP_CDR_ID_INDEX] = (char *)malloc(65*sizeof(char));
                        free(mode); 
                        cur_group++;
                    }
                }
            }
            else if(!xmlStrcasecmp(child->name, (xmlChar *)"event")) {
                char *ev_cause, *ev_time;
                char * pu_e164 = NULL, * pu_h323 = NULL, * pu_dn = NULL;
                char * ev_msg_value = xml_get_prop(child, "msg");
                event_msg = lookup_msg(ev_msg_value);
                if(event_msg == UNKNOWN_MSG || event_msg == INFO_TO_MSG || event_msg == INFO_FROM_MSG) {
                    free(ev_msg_value);
                    continue;
                }

                child_next = child->next;
                if(child_next && !xmlStrcasecmp(child_next->name, (xmlChar *)"event")) {
                    char *pickup_from = xml_get_prop(child_next, "msg");
                    // the setup msg before pickup-from contains call info of the DTMF pickup object
                    // this should be overridden by pickup-from to tag, which contains the original caller
                    if(child_next->children && !xmlStrcasecmp(child_next->children->name, (xmlChar *)"to") && !strcasecmp(pickup_from, "pickup-from")) {
                        pu_e164 = xml_get_prop(child_next->children, "e164");
                        pu_h323 = xml_get_prop(child_next->children, "h323");
                        pu_dn = xml_get_prop(child_next->children, "dn");
                    }
                    free(pickup_from);
                }

                events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_CAUSE_INDEX] = (char *)malloc(65);
                events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_TIME_INDEX] = (char *)malloc(65);
                events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_CDR_ID_INDEX] = (char *)malloc(65);
                events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_ORDER_INDEX] = (char *)malloc(65);

                ev_time = xml_get_prop(child, "time");
                cdr->ev_time[cur_event] = (ev_time == NULL) ? -1 : atoi(ev_time);
                free(ev_time);
                events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_MSG_INDEX] = ev_msg_value;
                if (!strcasecmp(ev_msg_value, "") || cdr->ev_time[cur_event] == -1) {
                    free_cdr_data(cdr_data);
                    return error(cgi, NULL, "CDR event attribute 'msg'/'time' is missing", "", query_string, ERROR_CDR_ATTRIBUTE_MSG_TIME_MISSING, 0, doc, ERROR_CDR_ATTRIBUTE_MSG_TIME_MISSING, 1, 1);
                }
                char * to_e164 = NULL, * to_h323 = NULL, * to_dn = NULL;
                if(child->children && !xmlStrcasecmp(child->children->name, (xmlChar *)"to")) {
                    to_e164 = xml_get_prop(child->children, "e164");
                    to_h323 = xml_get_prop(child->children, "h323");
                    to_dn = xml_get_prop(child->children, "dn");
                }
                events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_TYPE_INDEX] = xml_get_prop(child, "type");
                events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_CONF_INDEX] = xml_get_prop(child, "conf");
                ev_cause = xml_get_prop(child, "cause");
                cdr->ev_cause[cur_event] = (ev_cause == NULL) ? 0 : atoi(ev_cause);
                free(ev_cause);
                char * e164 = xml_get_prop(child, "e164");
                char * h323 = xml_get_prop(child, "h323");
                char * dn = xml_get_prop(child, "dn");
                int free_ev_vals = 1;
                int free_pu_vals = 1;
                if(pu_e164 && pu_h323 && pu_dn) { // not null
                    events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_H323_INDEX] = pu_h323;
                    events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_E164_INDEX] = pu_e164;
                    events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_DN_INDEX] = pu_dn;
                    free_pu_vals = 0;
                    if(to_e164) free(to_e164);
                    if(to_h323) free(to_h323);
                    if(to_dn)   free(to_dn);
                } else if((to_e164 && *to_e164) || (to_h323 && *to_h323) || (to_dn && *to_dn)) {
                    if(setup_event == cur_event && cur_event) {
                        if(!*(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_H323_INDEX]) && 
                           !*(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_E164_INDEX]) && 
                           !*(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_DN_INDEX])) {
                            free(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_H323_INDEX]);
                            free(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_E164_INDEX]);
                            free(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_DN_INDEX]);
                            events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_H323_INDEX] = h323;
                            events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_E164_INDEX] = e164;
                            events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_DN_INDEX] = dn;
                            free_ev_vals = 0;
                        }
                    }
                    events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_H323_INDEX] = to_h323;
                    events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_E164_INDEX] = to_e164;
                    events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_DN_INDEX] = to_dn;
                }
                else {
                    events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_H323_INDEX] = h323;
                    events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_E164_INDEX] = e164;
                    events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_DN_INDEX] = dn;
                    free_ev_vals = 0;
                    if(to_e164) free(to_e164);
                    if(to_h323) free(to_h323);
                    if(to_dn)   free(to_dn);
                }
                if(free_pu_vals == 1) {
                    if(pu_e164) free(pu_e164);
                    if(pu_h323) free(pu_h323);
                    if(pu_dn) free(pu_dn);
                }
                if(free_ev_vals == 1) {
                    if(e164) free(e164);
                    if(h323) free(h323);
                    if(dn) free(dn);
                }
                if(alert_time != -1) {
                    alert_duration = cdr->ev_time[cur_event] - alert_time;
                    alert_time = -1;
                }

                switch(event_msg) {
                    case SETUP_TO_MSG:
                    case SETUP_FROM_MSG:
                        cdr->cdrp_values[CDRP_CONF_INDEX] = str_malloc(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_CONF_INDEX]);
                        setup_event = cur_event + 1;
                        if(!*(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_H323_INDEX]) && 
                           !*(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_DN_INDEX]) && 
                            *(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_E164_INDEX])) {
                            cdr->search_ldap[cur_event] = 1;
                        }
                        break;
                    case ALERT_TO_MSG:
                    case ALERT_FROM_MSG:
                    case CONN_TO_MSG:
                    case CONN_FROM_MSG:
                    case CONNECTED_MSG:
                        if(event_msg == ALERT_TO_MSG || event_msg == ALERT_FROM_MSG) {
                            alert_time = cdr->ev_time[cur_event];
                        } else {
                            if(billing_call && billing_time == -1) billing_time = cdr->ev_time[cur_event];
                            if(call_time == -1) call_time = cdr->ev_time[cur_event];
                            if(conn_time == -1 && set_conn_time == 1) conn_time = cdr->ev_time[cur_event];
                        }
                        if(!*(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_H323_INDEX]) && 
                           !*(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_DN_INDEX]) && 
                            *(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_E164_INDEX])) {
                            cdr->search_ldap[cur_event] = 1;
                        }
                        break;
                    case DISC_TO_MSG:
                    case DISC_FROM_MSG:
                        disc_time = cdr->ev_time[cur_event];
                        if(billing_call && end_billing_time == -1) end_billing_time = cdr->ev_time[cur_event];
                        cause = cdr->ev_cause[cur_event];
                        break;
                    case REL_TO_MSG:
                    case REL_FROM_MSG:
                    {
                        char * more_calls = xml_get_prop(child, "more");
                        if (!strcasecmp(more_calls, "true")) {
                            if (cdr->cdrp_values[CDRP_MORE_CALLS_INDEX]) free(cdr->cdrp_values[CDRP_MORE_CALLS_INDEX]);
                            cdr->cdrp_values[CDRP_MORE_CALLS_INDEX] = str_malloc("t");
                        }
                        free(more_calls);
                        if (cause == -1) cause = cdr->ev_cause[cur_event];
                        if (billing_call && end_billing_time == -1) end_billing_time = cdr->ev_time[cur_event];
                    }   break;
                    case TRANSFER_TO_MSG:
                    case TRANSFER_FROM_MSG:
                    case TRANSFER_REMOTE:
                    case CF_TO_MSG:
                    case CF_FROM_MSG:
                        if(event_msg == TRANSFER_FROM_MSG || event_msg == CF_FROM_MSG) {
                            if(conn_time != -1 && conn_end_time == -1) 
                                conn_end_time = cdr->ev_time[cur_event];
                            set_conn_time = 0;
                        }
                        if(!*(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_H323_INDEX]) && 
                           !*(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_DN_INDEX]) && 
                            *(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_E164_INDEX])) {
                            cdr->search_ldap[cur_event] = 1;
                        }
                        // In this case, setup_from, cf-to, we check LDAP for an internal number
                        if((event_msg == CF_TO_MSG) && cur_event && (lookup_msg(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_MSG_INDEX]) == SETUP_FROM_MSG) && cdr->search_ldap[cur_event-1]) {
                            cdr->search_ldap[cur_event-1] = 2;
                        }
                        break;
                    case TRANSFERRED_MSG:
                        if(cdr->cdr_values[CDR_TRANSFERRED_INDEX]) free(cdr->cdr_values[CDR_TRANSFERRED_INDEX]);
                        cdr->cdr_values[CDR_TRANSFERRED_INDEX] = str_malloc("1");
                        break;
                    case FORWARDED_MSG:
                        if(cdr->cdr_values[CDR_FORWARDED_INDEX]) free(cdr->cdr_values[CDR_FORWARDED_INDEX]);
                        cdr->cdr_values[CDR_FORWARDED_INDEX] = str_malloc("1");
                        break;
                    case REJECT_MSG:
                        if (cdr->cdrp_values[CDRP_REJECTED_INDEX]) free(cdr->cdrp_values[CDRP_REJECTED_INDEX]);
                        cdr->cdrp_values[CDRP_REJECTED_INDEX] = str_malloc("t");
                        break;
                }
                time = cdr->ev_time[cur_event];
                if(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_TYPE_INDEX] && !strcasecmp(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_TYPE_INDEX], "ext")) {
                    free(cdr->cdrp_values[CDRP_TYPE_INDEX]);
                    cdr->cdrp_values[CDRP_TYPE_INDEX] = str_malloc("ext");
                }
                sprintf(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_ORDER_INDEX],"%d",cur_event);
                cur_event++;           
                // go through child "ep", "cf" and "ct"
                xmlNodePtr sub_child;
                char * new_pu_e164 = NULL, * new_pu_h323 = NULL, * new_pu_dn = NULL;
                int num_param_zero = 0;
                for(sub_child = child->children; sub_child; sub_child = sub_child->next) {
                    if(event_msg == PICKUP_FROM_MSG) {
                        char *pickup_from_se = xml_get_prop(sub_child, "msg");
                        if(!strcasecmp(pickup_from_se, "setup-from")) {
                            new_pu_e164 = xml_get_prop(sub_child, "e164");
                            new_pu_h323 = xml_get_prop(sub_child, "h323");
                            new_pu_dn = xml_get_prop(sub_child, "dn");
                            if(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_E164_INDEX]) {
                                free(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_E164_INDEX]);
                                events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_E164_INDEX] = NULL;
                            }
                            if(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_H323_INDEX]) {
                                free(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_H323_INDEX]);
                                events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_H323_INDEX] = NULL;
                            }
                            if(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_DN_INDEX]) {
                                free(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_DN_INDEX]);
                                events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_DN_INDEX] = NULL;
                            }
                            events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_E164_INDEX] = new_pu_e164;
                            events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_H323_INDEX] = new_pu_h323;
                            events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_DN_INDEX] = new_pu_dn;
                            if(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_CONF_INDEX]) 
                                free(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_CONF_INDEX]);
                            events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_CONF_INDEX] = xml_get_prop(sub_child, "conf");
                            num_param_zero = (!strcasecmp(new_pu_e164, "")) + (!strcasecmp(new_pu_h323, "")) + (!strcasecmp(new_pu_dn, "")); 
                            if(!*(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_H323_INDEX]) && 
                               !*(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_DN_INDEX]) && 
                                *(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_E164_INDEX])) {
                                cdr->search_ldap[cur_event-1] = 1;
                            }
                            else cdr->search_ldap[cur_event-1] = 0;
                        }
                        else if(!strcasecmp(pickup_from_se, "alert-to")) {
                            char *tnew_pu_e164 = xml_get_prop(sub_child, "e164");
                            char *tnew_pu_h323 = xml_get_prop(sub_child, "h323");
                            char *tnew_pu_dn = xml_get_prop(sub_child, "dn");
                            int num_param_zero_tmp = (!strcasecmp(tnew_pu_e164, "")) + (!strcasecmp(tnew_pu_h323, "")) + (!strcasecmp(tnew_pu_dn, ""));
                            if(num_param_zero_tmp < num_param_zero) {
                                if(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_E164_INDEX])
                                    free(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_E164_INDEX]);
                                events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_E164_INDEX] = tnew_pu_e164;
                                if(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_H323_INDEX])
                                    free(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_H323_INDEX]);
                                events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_H323_INDEX] = tnew_pu_h323;
                                if(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_DN_INDEX])
                                    free(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_DN_INDEX]);
                                events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_DN_INDEX] = tnew_pu_dn;
                                if(!*(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_H323_INDEX]) && 
                                   !*(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_DN_INDEX]) && 
                                    *(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_E164_INDEX])) {
                                    cdr->search_ldap[cur_event-1] = 1;
                                }
                                else cdr->search_ldap[cur_event-1] = 0;
                            }
                            else {
                                if(tnew_pu_e164) free(tnew_pu_e164);
                                if(tnew_pu_h323) free(tnew_pu_h323);
                                if(tnew_pu_dn)   free(tnew_pu_dn);
                            }
                        }
                        free(pickup_from_se);
                    }
                    else {
                        if(!xmlStrcasecmp(sub_child->name, (xmlChar *)"ep")) {
                            char * ep_e164 = xml_get_prop(sub_child, "e164");
                            char * ep_h323 = xml_get_prop(sub_child, "h323");
                            char * ep_dn = xml_get_prop(sub_child, "dn");
                            free(cdr->cdr_values[CDR_E164_INDEX]);
                            free(cdr->cdr_values[CDR_H323_INDEX]);
                            cdr->cdr_values[CDR_E164_INDEX] = (char *)ep_e164;
                            if(*ep_h323) {
                                cdr->cdr_values[CDR_H323_INDEX] = (char *)ep_h323;
                                free(ep_dn);
                            }
                            else {
                                free(ep_h323);
                                cdr->cdr_values[CDR_H323_INDEX] = (char *)ep_dn;
                            }
                        }
                        else if(sub_child->properties && (!xmlStrcasecmp(sub_child->name, (xmlChar *)"cf") || !xmlStrcasecmp(sub_child->name, (xmlChar *)"ct"))) {
                            events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_CAUSE_INDEX] = (char *)malloc(65);
                            events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_TIME_INDEX] = (char *)malloc(65);
                            events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_CDR_ID_INDEX] = (char *)malloc(65);
                            events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_ORDER_INDEX] = (char *)malloc(65);
                            events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_E164_INDEX] = xml_get_prop(sub_child, "e164");
                            events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_H323_INDEX] = xml_get_prop(sub_child, "h323");
                            events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_DN_INDEX] = xml_get_prop(sub_child, "dn");
                            if(!*(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_H323_INDEX]) && 
                               !*(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_DN_INDEX]) && 
                                *(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_E164_INDEX])) {
                                cdr->search_ldap[cur_event] = 1;
                            } else
                                cdr->search_ldap[cur_event] = 0;
                            cdr->ev_time[cur_event] = cdr->ev_time[cur_event - 1];
                            events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_TYPE_INDEX] = str_malloc(events_values[(cur_event-1)*EVENTS_NUM_FIELDS+EVENT_TYPE_INDEX]);
                            events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_CONF_INDEX] = str_malloc("");
                            cdr->ev_cause[cur_event] = 0;
                            if(!xmlStrcasecmp(sub_child->name, (xmlChar *)"cf")) {
                                events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_MSG_INDEX] = str_malloc("scf");
                            }
                            else {
                                events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_MSG_INDEX] = str_malloc("sct");
                            }
                            sprintf(events_values[cur_event*EVENTS_NUM_FIELDS+EVENT_ORDER_INDEX],"%d",cur_event);
                            cur_event++;
                        }
                    }
                }
            }
        }
        if(cause == -1) {
            free_cdr_data(cdr_data);
            return error(cgi, NULL, "Missing release/disconnect event in CDR", "", query_string, ERROR_CDR_REL_DISC_EVENT_MISSING, 0, doc, ERROR_CDR_REL_DISC_EVENT_MISSING, 1, 1);
        }
        if(conn_time == -1) conn_duration = 0;
        else
        {
            if(conn_end_time == -1) {
                if(disc_time != -1) conn_end_time = disc_time;
                else conn_end_time = time;
            }

            if(conn_end_time == conn_time)
                conn_duration = 1;
            else
                conn_duration = conn_end_time - conn_time;
        }
        if(billing_time == -1) billing_duration = 0;
        else 
        {
            if(end_billing_time == -1) end_billing_time = time;

            if(end_billing_time == billing_time)
                billing_duration = 1;
            else
                billing_duration = end_billing_time - billing_time;
        }
        if(call_time == -1) call_duration = 0;
        else
        {
            if(disc_time != -1) call_end_time = disc_time;
            else call_end_time = time;

            if(call_end_time == call_time)
                call_duration = 1;
            else
                call_duration = call_end_time - call_time;
        }

        if(strcmp(cdr->cdr_values[CDR_PHYS_INDEX], "") || strcmp(cdr->cdr_values[CDR_PBX_INDEX], "")) {
            read_ldap_config(NULL, cdr_data->ldap_mutex, cdr_data->ldap_act);
            pthread_mutex_lock(cdr_data->ldap_mutex);
            int ldap_init_ret = ldap_init(&ldap_conn, cdr->cdr_values[CDR_PHYS_INDEX], cdr->cdr_values[CDR_PBX_INDEX], *cdr_data->ldap_act);
            pthread_mutex_unlock(cdr_data->ldap_mutex);
            if(ldap_init_ret == 1) {
                do_ldap = 1;

                ldap_t **ptr_servers = ldap_conn.ldap_server;
                char *el_prefix = NULL;
                while(*ptr_servers != NULL) {
                    if(strncmp(cdr->cdr_values[CDR_E164_INDEX], (*ptr_servers)->el, strlen((*ptr_servers)->el)) == 0) {
                        el_prefix = (*ptr_servers)->el;
                        break;
                    }
                    ptr_servers++;
                }
                // LDAP for the cdr if external number
                if(el_prefix && cdr->cdr_values[CDR_E164_INDEX]) {
                    cdr->cdr_values[CDR_DN_ESCAPED_INDEX] = ldap_get_name(&ldap_conn, cdr->cdr_values[CDR_E164_INDEX], 0);
                }
            }
            else if(ldap_init_ret == 2) {
                free_cdr_data(cdr_data);
                return error(cgi, NULL, "LDAP servers couldn't be read", "", query_string, ERROR_LDAP_SERVER_READ, 0, doc, ERROR_LDAP_SERVER_READ, 1, 0);
            }
        }
        clear_alarm(cgi, ERROR_LDAP_SERVER_READ);

        cdrp_flow_size = 3 + cdrp_flow_seperators + strlen(cdr->cdr_values[CDR_DIR_INDEX]) + strlen(cdr->cdr_values[CDR_E164_INDEX]) + strlen(cdr->cdr_values[CDR_CN_INDEX]); 
        if(cdr->cdr_values[CDR_DN_ESCAPED_INDEX] == NULL) {
            cdr->cdr_values[CDR_DN_ESCAPED_INDEX] = str_malloc("");
            cdrp_flow_size += strlen(cdr->cdr_values[CDR_H323_INDEX]);
        }
        else
            cdrp_flow_size += strlen(cdr->cdr_values[CDR_DN_ESCAPED_INDEX]);

        for(i=0; i<num_events; i++) {
            cdr->ev_h323_escaped[i] = inno_escape(events_values[i*EVENTS_NUM_FIELDS+EVENT_H323_INDEX]);
            if(do_ldap && cdr->search_ldap[i]) {
                // we look the previous events to avoid asking for the same number
                for(j = 0; j < i; j++) {
                    if(!strcmp(events_values[j*EVENTS_NUM_FIELDS+EVENT_E164_INDEX], events_values[i*EVENTS_NUM_FIELDS+EVENT_E164_INDEX]) && (cdr->ev_dn_escaped[j] != NULL)) {
                        cdr->ev_dn_escaped[i] = malloc(strlen(cdr->ev_dn_escaped[j])+1);
                        strcpy(cdr->ev_dn_escaped[i], cdr->ev_dn_escaped[j]);
                        cdr->search_ldap[i] = 0;
                        break;
                    }
                }
                char *t_dn = ldap_get_name(&ldap_conn, events_values[i*EVENTS_NUM_FIELDS+EVENT_E164_INDEX], cdr->search_ldap[i]);
                if(t_dn == NULL) {
                    if(cdr->ev_dn_escaped[i] == NULL) cdr->ev_dn_escaped[i] = inno_escape(events_values[i*EVENTS_NUM_FIELDS+EVENT_DN_INDEX]);
                }
                else {
                    if(cdr->ev_dn_escaped[i]) free(cdr->ev_dn_escaped[i]);
                    cdr->ev_dn_escaped[i] = t_dn;
                }
                if(trace_level & TRACE_STD) {
                    do_log(LOG_REPORTING, "", "search number[%d] number = %s message = %s Name = %s", cdr->search_ldap[i], events_values[i*EVENTS_NUM_FIELDS+EVENT_E164_INDEX], events_values[i*EVENTS_NUM_FIELDS+EVENT_MSG_INDEX], cdr->ev_dn_escaped[i]);
                }
            }
            else {
                cdr->ev_dn_escaped[i] = inno_escape(events_values[i*EVENTS_NUM_FIELDS+EVENT_DN_INDEX]);
            }
            cdrp_flow_size += cdrp_flow_seperators + strlen(events_values[i*EVENTS_NUM_FIELDS+EVENT_MSG_INDEX]) + strlen(events_values[i*EVENTS_NUM_FIELDS+EVENT_E164_INDEX]) + strlen(cdr->ev_h323_escaped[i]) + strlen(cdr->ev_dn_escaped[i]);
        }

        if(do_ldap) ldap_close(&ldap_conn);

        // start cdrp call flow
        int cdrp_flow_len;
        char *cdrp_flow = malloc(cdrp_flow_size);
        if(!strcmp(cdr->cdr_values[CDR_DN_ESCAPED_INDEX], "")) {
            cdrp_flow_len = sprintf(cdrp_flow, "{{%s,%s,%s,%s},", cdr->cdr_values[CDR_DIR_INDEX], cdr->cdr_values[CDR_E164_INDEX], cdr->cdr_values[CDR_H323_INDEX], cdr->cdr_values[CDR_CN_INDEX]);
        }
        else {
            cdrp_flow_len = sprintf(cdrp_flow, "{{%s,%s,%s,},", cdr->cdr_values[CDR_DIR_INDEX], cdr->cdr_values[CDR_E164_INDEX], cdr->cdr_values[CDR_DN_ESCAPED_INDEX]);
        }

        if(num_events > 5) {
            events_query = malloc(strlen(PG_EVENTS_QUERY_BEGIN) + (num_events * 60));
            int ev_ins_bytes = sprintf(events_query, "%s", PG_EVENTS_QUERY_BEGIN);
            index_arg = 1;
            for(i=0; i<num_events; i++) {
                sprintf(events_values[i*EVENTS_NUM_FIELDS+EVENT_CAUSE_INDEX], "%d", cdr->ev_cause[i]);
                sprintf(events_values[i*EVENTS_NUM_FIELDS+EVENT_TIME_INDEX], "%d", cdr->ev_time[i]);

                cdrp_flow_len += sprintf(&cdrp_flow[cdrp_flow_len], "{%s,%s,%s,%s},", events_values[i*EVENTS_NUM_FIELDS+EVENT_MSG_INDEX], events_values[i*EVENTS_NUM_FIELDS+EVENT_E164_INDEX], cdr->ev_h323_escaped[i], cdr->ev_dn_escaped[i]);

                ev_ins_bytes += sprintf(&events_query[ev_ins_bytes], "($%d,$%d,$%d,$%d,$%d,$%d,$%d,$%d,$%d,$%d),",
                                index_arg, index_arg+1, index_arg+2, index_arg+3, index_arg+4,
                                index_arg+5, index_arg+6, index_arg+7, index_arg+8, index_arg+9);
                index_arg += EVENTS_NUM_FIELDS;
            }
            events_query[ev_ins_bytes - 1] = '\0';
            cdrp_flow[cdrp_flow_len - 1] = '}'; // overwrite last , with }
        }
        else if(num_events > 0) {
            events_query = strdup(cdr_data->ev_prep_query[num_events-1]);
            for(i=0; i<num_events; i++) {
                sprintf(events_values[i*EVENTS_NUM_FIELDS+EVENT_CAUSE_INDEX], "%d", cdr->ev_cause[i]);
                sprintf(events_values[i*EVENTS_NUM_FIELDS+EVENT_TIME_INDEX], "%d", cdr->ev_time[i]);

                cdrp_flow_len += sprintf(&cdrp_flow[cdrp_flow_len], "{%s,%s,%s,%s},", events_values[i*EVENTS_NUM_FIELDS+EVENT_MSG_INDEX], events_values[i*EVENTS_NUM_FIELDS+EVENT_E164_INDEX], cdr->ev_h323_escaped[i], cdr->ev_dn_escaped[i]);
            }
            cdrp_flow[cdrp_flow_len - 1] = '}'; // overwrite last , with }
        }
        else {
            sprintf(cdrp_flow, "{}");
        }
        if(trace_level & TRACE_DB) {
            do_log(LOG_REPORTING, "", "cdrp flow[%d]=%s query=%s", num_events, cdrp_flow, events_query);
        }

        // start cdrp->groups
        char *cdrp_groups = malloc(cdrp_groups_size);
        int cdrp_groups_len = sprintf(cdrp_groups, "{");

        if(num_groups > 5) {
            groups_query = malloc(strlen(PG_GROUPS_QUERY_BEGIN) + (num_groups * 30));
            int grp_ins_bytes = sprintf(groups_query, "%s", PG_GROUPS_QUERY_BEGIN);
            index_arg = 1;
            for(i=0; i<num_groups; i++) {
                grp_ins_bytes += sprintf(&groups_query[grp_ins_bytes], "($%d,$%d,$%d,$%d),", index_arg, index_arg+1, index_arg+2, index_arg+3);
                index_arg += GROUPS_NUM_FIELDS;

                cdrp_groups_len += sprintf(&cdrp_groups[cdrp_groups_len], "{%s},", cdr->groups_name_escaped[i]);
            }
            groups_query[grp_ins_bytes - 1] = '\0';
            cdrp_groups[cdrp_groups_len - 1] = '}'; //overwrite last , with }
        } 
        else if(num_groups > 0) {
            groups_query = strdup(cdr_data->grp_prep_query[num_groups-1]);
            for(i=0; i<num_groups; i++) {
                cdrp_groups_len += sprintf(&cdrp_groups[cdrp_groups_len], "{%s},", cdr->groups_name_escaped[i]);
            }
            cdrp_groups[cdrp_groups_len - 1] = '}'; //overwrite last , with }
        }
        else {
            sprintf(cdrp_groups, "{}");
        }

        if(trace_level & TRACE_DB) {
            do_log(LOG_REPORTING, "", "cdrp groups[%d]=%s query=%s", num_groups, cdrp_groups, groups_query);
        }

        cdr->cdrp_values[CDRP_CDR_ID_INDEX] = (char *)malloc(32*sizeof(char));
        cdr->cdrp_values[CDRP_CALL_FLOW_INDEX] = cdrp_flow;
        cdr->cdrp_values[CDRP_GROUPS_INDEX] = cdrp_groups;
        cdr->cdrp_values[CDRP_CONN_DUR_INDEX] = (char *)malloc(32);
        sprintf(cdr->cdrp_values[CDRP_CONN_DUR_INDEX],"%d",conn_duration); 
        cdr->cdrp_values[CDRP_BILL_DUR_INDEX] = (char *)malloc(32);
        sprintf(cdr->cdrp_values[CDRP_BILL_DUR_INDEX],"%d",billing_duration); 
        cdr->cdrp_values[CDRP_CALL_DUR_INDEX] = (char *)malloc(32);
        sprintf(cdr->cdrp_values[CDRP_CALL_DUR_INDEX],"%d",call_duration); 
        cdr->cdrp_values[CDRP_ALERT_DUR_INDEX] = (char *)malloc(32);
        sprintf(cdr->cdrp_values[CDRP_ALERT_DUR_INDEX],"%d",alert_duration); 
        cdr->cdrp_values[CDRP_CAUSE_INDEX] = (char *)malloc(32);
        sprintf(cdr->cdrp_values[CDRP_CAUSE_INDEX],"%d",cause); 

        cdr->groups_query = groups_query;
        cdr->events_query = events_query;

        cdr_index++;
    }
    xmlFreeDoc(doc);

    if(trace_level & TRACE_TIME) {
        gettimeofday(&tvEnd, NULL);
        timeval_subtract(&tvDiff, &tvEnd, &tvBegin);
        do_log(LOG_REPORTING, "", "cdr proc: %ld.%06ld", tvDiff.tv_sec, tvDiff.tv_usec);
    }

    //syslog(LOG_INFO, "DONE");
    if(trace_level & TRACE_STD) {
        system("echo 'Test-CDR-End:' >> /tmp/mem.log");
        system("ps -eo pmem,pcpu,rss,vsize,args | sort -k 1 | grep 'process.fcgi' | grep -v grep >> /tmp/mem.log");
    }
    if(send_err) return send_err;
    return 0;
}
